package net.sergeych.crypto2

import kotlinx.serialization.SerialName
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient

@Serializable
sealed class UniversalKey {
    abstract val id: KeyId

    @Serializable
    @SerialName("sym")
    data class Symmetric(val key: SymmetricKey) : UniversalKey(), EncryptingKey by key, DecryptingKey by key {
        @Transient
        override val id: KeyId = key.id

        @Transient
        override val nonceBytesLength: Int = key.nonceBytesLength

        override fun toString() = "U.Sym:$id"
    }

    @Serializable
    @SerialName("ssn")
    data class Session(val key: SafeKeyExchange.SessionKey) : UniversalKey(), EncryptingKey by key,
        DecryptingKey by key {
        @Transient
        override val id: KeyId = key.id
        @Transient
        override val nonceBytesLength: Int = key.nonceBytesLength
    }

    @Serializable
    @SerialName("sec")
    data class Secret(val key: Asymmetric.SecretKey) : UniversalKey(), DecryptingKey by key {

        @Transient
        val publicKey: Public = Public(key.publicKey)

        override val id: KeyId by lazy { key.id }
        override fun toString() = "U.Sec:$id"
    }

    @Serializable
    @SerialName("pub")
    data class Public(val key: Asymmetric.PublicKey) : UniversalKey(), EncryptingKey {

        override val id: KeyId by lazy { key.id }

        override fun encryptWithNonce(plainData: UByteArray, nonce: UByteArray, randomFill: IntRange?): UByteArray =
            key.encryptMessage(plainData, nonce, randomFill= randomFill).encoded

        override fun encrypt(plainData: UByteArray, randomFill: IntRange?): UByteArray =
            key.encryptMessage(plainData, randomFill=randomFill).encoded

        override val nonceBytesLength: Int
            get() = Asymmetric.nonceBytesLength

        override fun toString() = "U.Sec:$id"
    }

    @Serializable
    @SerialName("sig")
    data class Signing(val key: SigningSecretKey) : UniversalKey(), SigningKey by key {
        override val id: KeyId by lazy { key.id }
        override fun toString() = "U.Sig:$id"

        /**
         * [Verifying] key, e.g. [verifyingKey] wrapped in the [UniversalKey] variant.
         */
        val publicKey by lazy { Verifying(verifyingKey) }

    }

    @Serializable
    @SerialName("ver")
    data class Verifying(val key: SigningPublicKey) : UniversalKey(), VerifyingKey by key {
        override val id: KeyId by lazy { key.id }
        override fun toString() = "U.Sig:$id"
    }



    companion object {
        fun from(key: DecryptingKey): UniversalKey =
            when (key) {
                is UniversalKey -> key
                is Asymmetric.SecretKey -> Secret(key)
                is SymmetricKey -> Symmetric(key)
                is SafeKeyExchange.SessionKey -> Session(key)
                else -> throw UnsupportedOperationException("can't create universal key from ${key::class.simpleName}")
            }

        fun newSecretKey(): Secret =
            Secret(Asymmetric.new())
        fun newSigningKey(): Signing =
            Signing(SigningSecretKey.new())
        @Suppress("unused")
        fun newSymmetricKey(): Symmetric =
            Symmetric(SymmetricKey.new())
    }

}
